/**********************************************************************\
**
**                         Copyright (c) 2009-2018 by
**                          HENSGSTLER GmbH, Aldingen
**
**      This  software  is furnished under a licence and may be used and
**      copied only in accordance with the terms  of  such  licence  and
**      with the inclusion of the above copyright notice.  This software
**      or any other copies thereof may not  be  provided  or  otherwise
**      made  available  to any other person.  No title to and ownership
**      of the software is hereby transferred.
**
**      The  information  in  this software is subject to change without
**      notice and should not be construed as a commitment by Hengstler.
**
**  Module:      Exo_Api.h
**  Facility:    Hengstler Extendo Linux Printer Driver Package / S684018
**  Application: Extendo application programming interface (API) for linux (libExoApi)
**  Version:     R3-V1_04e-BETA-2018AUG14
**  Environment: Linux Debian
**  Compiler:    gcc
**  Authors:     L. Tkac
**  Abstract:    Api Library header file
**  History:     R2-V1.01  / 2009JUN16 / TKA / Initial version
**               R3-V1.02a / 2011DEC14 / TKA / TKA-COR-03-New API communication added (libusb)
**               R3-V1.04a / 2018MAR23 / TKA / USB 2.0 support added
**                                             API extended for standard macros   
**               R3-V1.04b / 2018JUN25 / TKA / Implemented intermediate read buffer(FIFO)
**                                             Fix for reading separate header + body data packages   
**                                             Added ERROR reading codes   
**               R3-V1.04e / 2018AUG14 / TKA / Return value of printer handle change from 'int' to 'intptr_t' 
**                                             due to 32/64bit incompatibilities on some 64bit distributions (Ubuntu 18.04-64bit)
**
\**********************************************************************/

#ifndef __EXO_API_H
#define __EXO_API_H

/*
** INCLUDES
*/

#ifdef __cplusplus
extern "C" {
#endif
   
#include <stdint.h>
   

/*
** MACROS
*/

//-- Error codes of application programming interface
#define D_EXO_API_OPEN_FAILED                                              -1
#define D_EXO_API_INVALID_DEVICE                                           -2
#define D_EXO_API_DETACH_FAILED                                            -3
#define D_EXO_API_CLAIM_FAILED                                             -4
#define D_EXO_API_GET_ACTIVE_CONFIG_FAILED                                 -5
#define D_EXO_API_NO_INTERFACE                                             -6
#define D_EXO_API_ENDPOINT_COUNT_FAILURE                                   -7
#define D_EXO_API_COMM_BUFFER_FAILURE                                      -8

//-- Error codes of "exo_api_printer_read" interface
#define D_EXO_API_READ_BUFFER_FAILURE                                      -1
#define D_EXO_API_READ_SAVING_FAILURE                                      -2
#define D_EXO_API_READ_OVERFLOW                                            -3

//-- Extendo printer access options
#define D_EXO_API_PIO_OPT_NONE                                                0x00000000 // Option indicating NO OPTIONS                      
#define D_EXO_API_PIO_OPT_NO_SYNCHRONIZE                                      0x00000001 // Option indicating NO SYNCHRONIZE on connect       

//-- Extendo printer access error codes from GetLastError()
#define D_EXO_API_PIO_ERR_OUTOFMEMORY                                         0x40000001 // Error code indicating out of memory               
#define D_EXO_API_PIO_ERR_STOPSPOOLERFAILED                                   0x40000002 // Error code indicating stop spooler failed         
#define D_EXO_API_PIO_ERR_INVALIDPARAMETER                                    0x40000003 // Error code indicating invalid parameter           
#define D_EXO_API_PIO_ERR_ENDOFENUM                                           0x40000004 // Error code indicating end of enumeration          
#define D_EXO_API_PIO_ERR_OPENPRINTERFAILED                                   0x40000005 // Error code indicating open printer failed         
#define D_EXO_API_PIO_ERR_READPRINTERFAILED                                   0x40000006 // Error code indicating read printer failed         
#define D_EXO_API_PIO_ERR_WRITEPRINTERFAILED                                  0x40000007 // Error code indicating write printer failed        
#define D_EXO_API_PIO_ERR_UNKNOWNPRINTER                                      0x40000008 // Error code indicating unknown printer             

//-- Maximum values
#define D_EXO_API_PIO_S_PRINTER_NAME                                          200        // Maximum length of printer name (excl. zero)       
#define D_EXO_API_PIO_S_COMMAND                                               2000       // Maximum length of command (excl. zero)            

//-- Protocol revision 
#define D_EXO_API_PCKT_REVISION                                               0          // Revision of the data packets

//-- Data request id's which correspond with bits in parameter [n] in emulation command "GS A+[n]"
//.. After reception of this command, the printer will send a sequence of requested data packets
//.. The packet data "[d1]+[dn]" immediately follows the packet header.
#define D_EXO_API_DATA_REQUEST_ID_PRINTER_STATUS                              0x01       // Bit number for inquiring printer status
#define D_EXO_API_DATA_REQUEST_ID_PRINTER_INFO                                0x02       // Bit number for inquiring printer info
#define D_EXO_API_DATA_REQUEST_ID_PRINTER_SENSORS                             0x04       // Bit number for inquiring sensor status
#define D_EXO_API_DATA_REQUEST_ID_PRINTER_FLASH                               0x08       // Bit number for inquiring flash contents
#define D_EXO_API_DATA_REQUEST_ID_PRESENTER_STATUS                            0x10       // Bit number for inquiring presenter status (Available from firmware version R4-2.00)
#define D_EXO_API_DATA_REQUEST_ID_PRINTER_LIFE_CYCLE                          0x20       // Bit number for inquiring life cycle data (Available from firmware version R4-V1.03b)

//-- Data packets id's which correspond with parameter ubId in TS_EXO_API_PCKT_HEADER
//.. Note that the packet header "ESC [FF]+[i]+[n]" has [i] as packet identifier and [n] as packet data size.
//.. The packet data "[d1]+[dn]" immediately follows the packet header.
#define D_EXO_API_DATA_PACKET_ID_PRINTER_STATUS                               0x02       // Packet identifier of a printer status packet
#define D_EXO_API_DATA_PACKET_ID_PRINTER_INFO                                 0x03       // Packet identifier of a printer info packet
#define D_EXO_API_DATA_PACKET_ID_PRINTER_SENSORS                              0x04       // Packet identifier of a sensor status packet
#define D_EXO_API_DATA_PACKET_ID_PRINTER_FLASH                                0x05       // Packet identifier of a flash contents packet
#define D_EXO_API_DATA_PACKET_ID_PRESENTER_STATUS                             0x06       // Packet identifier of a presenter status packet (Available from firmware version R4-2.00)
#define D_EXO_API_DATA_PACKET_ID_PRINTER_LIFE_CYCLE                           0x07       // Packet identifier of a life cycle packet (Available from firmware version R4-V1.03b)

//-- Printer status packet -> printer status dword bits
#define D_EXO_API_PRINTER_STATUS_PSDW_PRINTER_ERROR                           0x00000001  
#define D_EXO_API_PRINTER_STATUS_PSDW_PRINT_HEAD_TEMP_ALERT                   0x00000002
#define D_EXO_API_PRINTER_STATUS_PSDW_FLASH_CHECKSUM_ERROR                    0x00000004
#define D_EXO_API_PRINTER_STATUS_PSDW_PRINTER_STALLED                         0x00000008
#define D_EXO_API_PRINTER_STATUS_PSDW_PAPER_OUT_DETECTED                      0x00000010
#define D_EXO_API_PRINTER_STATUS_PSDW_PAPER_LOW_DETECTED                      0x00000020
#define D_EXO_API_PRINTER_STATUS_PSDW_PAPER_MARK_DETECTED                     0x00000040
#define D_EXO_API_PRINTER_STATUS_PSDW_PAPER_IN_CHUTE_DETECTED                 0x00000080
#define D_EXO_API_PRINTER_STATUS_PSDW_PRINT_HEAD_UP_DETECTED                  0x00000100
#define D_EXO_API_PRINTER_STATUS_PSDW_LAST_PRINTOUT_LOST                      0x00000200
#define D_EXO_API_PRINTER_STATUS_PSDW_AUXILIARY_SENSOR_ACTIVE                 0x00000400
#define D_EXO_API_PRINTER_STATUS_PSDW_SYSTEM_FAILURE                          0x00000800
#define D_EXO_API_PRINTER_STATUS_PSDW_POWER_FAILURE                           0x00001000
#define D_EXO_API_PRINTER_STATUS_PSDW_CUSTOMER_FLAG_1                         0x00002000
#define D_EXO_API_PRINTER_STATUS_PSDW_CUSTOMER_FLAG_2                         0x00004000
#define D_EXO_API_PRINTER_STATUS_PSDW_PRINTER_NOT_READY                       0x00008000
#define D_EXO_API_PRINTER_STATUS_PSDW_PRINTER_BUSY                            0x00010000   // Available from firmware version R4-V1.04h
#define D_EXO_API_PRINTER_STATUS_PSDW_EEPROM_MISSING                          0x00020000   // Available from firmware version R3-V1.02g
#define D_EXO_API_PRINTER_STATUS_PSDW_PRESENTER_AVAILABLE                     0x00040000   // Available from firmware version R4-V1.04h
#define D_EXO_API_PRINTER_STATUS_PSDW_PRESENTER_ACTIVE                        0x00080000   // Available from firmware version R4-V1.04h
#define D_EXO_API_PRINTER_STATUS_PSDW_PARCUT_HOLDING_PRINTOUT                 0x00300000   // Available from firmware version R3-V1.02t
#define D_EXO_API_PRINTER_STATUS_PSDW_PARCUT_PRINTOUT_TAKEN                   0x00100000   // Available from firmware version R3-V1.02t
#define D_EXO_API_PRINTER_STATUS_PSDW_PARCUT_PRINTOUT_REMOVED                 0x00200000   // Available from firmware version R4-V1.02t

//-- Printer status packet -> Printer error word codes (high byte is WHAT, low byte is WHEN)
#define D_EXO_API_PRINTER_STATUS_PEW_MSB_WHAT_PAPER_JAMMED                    0x01  
#define D_EXO_API_PRINTER_STATUS_PEW_MSB_WHAT_CUTTER_JAMMED                   0x02  
#define D_EXO_API_PRINTER_STATUS_PEW_MSB_WHAT_PRESENTER_JAMMED                0x03
#define D_EXO_API_PRINTER_STATUS_PEW_MSB_WHAT_ERROR(ERR,WHAT)                 (((ERR)&0xFF00) == (((WHAT) & 0x00FF) << 8))
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_DURING_PRINT                    0x01
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_DURING_PAPER_LOAD               0x02
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_DURING_PAPER_END                0x03
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_DURING_INTIALIZATION            0x04
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_DURING_END_OF_PRINT             0x05
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_WHILE_IDLE                      0x06
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_WHILE_LOADING                   0x07 
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_WHILE_PRESENTING                0x08
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_WHILE_REJECTING                 0x09
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_WHILE_CANCELING                 0x0A
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_DURING_EJECT                    0x0B
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_DURING_TEST                     0x0C
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_WHILE_COUPLING                  0x0D
#define D_EXO_API_PRINTER_STATUS_PEW_LSB_WHEN_ERROR(ERR,WHEN)                 (((ERR)&0x00FF) == ((WHEN) & 0x00FF))

//-- Printer sensor packet -> printer sensor state (high nibble is type, low nibble is state)
#define D_EXO_API_PRINTER_SENSOR_PSS_MSN_NONE                                 0x00
#define D_EXO_API_PRINTER_SENSOR_PSS_MSN_REFLEX                               0x10
#define D_EXO_API_PRINTER_SENSOR_PSS_MSN_TLIGHT                               0x20
#define D_EXO_API_PRINTER_SENSOR_PSS_MSN_SWITCH                               0x30
#define D_EXO_API_PRINTER_SENSOR_PSS_LSN_NONE                                 0x00
#define D_EXO_API_PRINTER_SENSOR_PSS_LSN_NO_PAPER                             0x01
#define D_EXO_API_PRINTER_SENSOR_PSS_LSN_PAPER                                0x02
#define D_EXO_API_PRINTER_SENSOR_PSS_LSN_PARKED                               0x03
#define D_EXO_API_PRINTER_SENSOR_PSS_LSN_CUTTING                              0x04
#define D_EXO_API_PRINTER_SENSOR_PSS_LSN_HEAD_DOWN                            0x05
#define D_EXO_API_PRINTER_SENSOR_PSS_LSN_HEAD_UP                              0x06
#define D_EXO_API_PRINTER_SENSOR_PSS_LSN_ACTIVE                               0x07
#define D_EXO_API_PRINTER_SENSOR_PSS_LSN_INACTIVE                             0x08

//-- Printer flash packet -> printer flash area index/state 
//.. Extendo ...
#define D_EXO_API_PRINTER_FLASH_FA_INDEX_FIRMWARE                             0x00
#define D_EXO_API_PRINTER_FLASH_FA_INDEX_FONT_1                               0x01
#define D_EXO_API_PRINTER_FLASH_FA_INDEX_LOGO                                 0x02
#define D_EXO_API_PRINTER_FLASH_FA_INDEX_FONT_2                               0x03
#define D_EXO_API_PRINTER_FLASH_FA_INDEX_WIDE_FONT                            0x04
#define D_EXO_API_PRINTER_FLASH_FA_INDEX_IMAGES                               0x05
//.. Extendo-XF...
#define D_EXF_API_PRINTER_FLASH_FA_INDEX_FIRMWARE                             0x00
#define D_EXF_API_PRINTER_FLASH_FA_INDEX_STANDARD_FONT                        0x01
#define D_EXF_API_PRINTER_FLASH_FA_INDEX_LOGO                                 0x02
#define D_EXF_API_PRINTER_FLASH_FA_INDEX_FACTORY_FONT                         0x03
#define D_EXF_API_PRINTER_FLASH_FA_INDEX_CUSTOMER_FONT                        0x04
#define D_EXF_API_PRINTER_FLASH_FA_INDEX_CUSTOMER_IMAGES                      0x05
//.. Status
#define D_EXO_API_PRINTER_FLASH_FA_STATE_UNDEFINED                            0x00
#define D_EXO_API_PRINTER_FLASH_FA_STATE_OK                                   0x01
#define D_EXO_API_PRINTER_FLASH_FA_STATE_ERROR_CRC                            0x02
#define D_EXO_API_PRINTER_FLASH_FA_STATE_ERROR_TYPE                           0x03
#define D_EXO_API_PRINTER_FLASH_FA_STATE_ERROR_SIZE                           0x04

//-- Presenter status packet -> presenter status word
#define D_EXO_API_PRESENTER_STATUS_PSW_PRESENTER_READY                        0x0001
#define D_EXO_API_PRESENTER_STATUS_PSW_PRESENTER_BUSY                         0x0002
#define D_EXO_API_PRESENTER_STATUS_PSW_PRESENTER_OCCUPIED                     0x0004
#define D_EXO_API_PRESENTER_STATUS_PSW_PRINTOUTS_LOADED                       0x0008
#define D_EXO_API_PRESENTER_STATUS_PSW_PRESENTING_PRINTOUTS                   0x0010
#define D_EXO_API_PRESENTER_STATUS_PSW_REJECTING_PRINTOUTS                    0x0020
#define D_EXO_API_PRESENTER_STATUS_PSW_PRINTOUTS_TAKEN                        0x0040
#define D_EXO_API_PRESENTER_STATUS_PSW_PRINTOUTS_REJECTED                     0x0080
#define D_EXO_API_PRESENTER_STATUS_PSW_PRESENTER_JAMMED                       0x0100
#define D_EXO_API_PRESENTER_STATUS_PSW_PAPER_IN_CHUTE_DETECTED                0x0200

//-- Configuration data - customer sensors - paper pre end sensor - enumeration
#define D_EXO_API_CD_CS_PPES_MASK                                             0x7000
#define D_EXO_API_CD_CS_PPES_NONE                                             0x0000
#define D_EXO_API_CD_CS_PPES_REFLEX                                           0x1000
#define D_EXO_API_CD_CS_PPES_TLIGHT                                           0x2000
#define D_EXO_API_CD_CS_PPES_SWITCH                                           0x3000
#define D_EXO_API_CD_CS_PPES_DETECT_PAPER_WHEN_LOW                            0x8000

//-- Configuration data - customer sensors - auxiliary sensor - enumeration
#define D_EXO_API_CD_CS_AUXS_MASK                                             0x0700
#define D_EXO_API_CD_CS_AUXS_NONE                                             0x0000
#define D_EXO_API_CD_CS_AUXS_REFLEX                                           0x0100
#define D_EXO_API_CD_CS_AUXS_TLIGHT                                           0x0200
#define D_EXO_API_CD_CS_AUXS_SWITCH                                           0x0300
#define D_EXO_API_CD_CS_AUXS_ACTIVE_WHEN_LOW                                  0x0800

//-- Configuration data - customer sensors - reserved
#define D_EXO_API_CD_CS_RES_MASK                                              0x00FF

//-- Configuration data - UART Baudrates  
#define D_EXO_API_CD_RS232_BAUDRATE_4800                                      0
#define D_EXO_API_CD_RS232_BAUDRATE_9600                                      1
#define D_EXO_API_CD_RS232_BAUDRATE_19200                                     2
#define D_EXO_API_CD_RS232_BAUDRATE_38400                                     3
#define D_EXO_API_CD_RS232_BAUDRATE_57600                                     4
#define D_EXO_API_CD_RS232_BAUDRATE_115200                                    5
#define D_EXO_API_CD_RS232_BAUDRATE_230400                                    6 // Not supported 
#define D_EXO_API_CD_RS232_BAUDRATE_460800                                    7 // Not supported 
#define D_EXO_API_CD_RS232_BAUDRATE_MAX                                       D_EXO_API_CD_RS232_BAUDRATE_115200
#define D_EXO_API_CD_RS232_BAUDRATE_DEF                                       D_EXO_API_CD_RS232_BAUDRATE_115200

//-- Configuration data - UART parity on/off
#define D_EXO_API_CD_RS232_PARITY_OFF                                         0
#define D_EXO_API_CD_RS232_PARITY_ON                                          1

//-- Configuration data - UART parity odd/even (if on)    
#define D_EXO_API_CD_RS232_PARITY_ODD                                         0
#define D_EXO_API_CD_RS232_PARITY_EVEN                                        1

//-- Configuration data - UART data bits (7/8)
#define D_EXO_API_CD_RS232_DATABIT_7                                          0 // Not supported 
#define D_EXO_API_CD_RS232_DATABIT_8                                          1

//-- Configuration data - UART stop bits (1/2)
#define D_EXO_API_CD_RS232_STOPBIT_1                                          0
#define D_EXO_API_CD_RS232_STOPBIT_2                                          1

//-- Configuration data - UART flow control (handshake)
#define D_EXO_API_CD_RS232_FLOWCNTL_NONE                                      0 // Not supported 
#define D_EXO_API_CD_RS232_FLOWCNTL_HW                                        1
#define D_EXO_API_CD_RS232_FLOWCNTL_SW                                        2 // Not supported 

//-- Configuration data - Wide font setting (Extendo Firmware S684014)
#define D_EXO_API_CD_WF_NONE                                                  0 // Wide font disabled 
#define D_EXO_API_CD_WF_24X24_STANDARD                                        1 // Standard wide font      24X24 
#define D_EXO_API_CD_WF_24X24_GB3212                                          2 // GB2312 china wide font  24X24 
#define D_EXO_API_CD_WF_20X20_STANDARD                                        3 // Standard wide font      20X20 
#define D_EXO_API_CD_WF_20X20_KS5601                                          4 // KS5601 korean wide font 20X20 

//-- Configuration data - Options (Extendo Firmware S684014)
#define D_EXO_API_CD_OPTION_CUTTER_INIT_ON_POWER_ON                           0x00000001
#define D_EXO_API_CD_OPTION_MARK_FEED_BEFORE_CUT                              0x00000002
#define D_EXO_API_CD_OPTION_PRINT_NO_INFO_ON_LOAD                             0x00000004
#define D_EXO_API_CD_OPTION_DISABLE_CUT_ON_LOAD                               0x00000008
#define D_EXO_API_CD_OPTION_INVALIDATE_PAPER_ON_LOAD                          0x00000010
#define D_EXO_API_CD_OPTION_NO_PAPER_FEED_AFTER_LOAD                          0x00000020
#define D_EXO_API_CD_OPTION_DISABLE_RETRACT_ON_PAPER_END                      0x00000040
#define D_EXO_API_CD_OPTION_CONTINUE_PRINT_ON_PAPER_END                       0x00000080
#define D_EXO_API_CD_OPTION_INVALIDATE_PAPER_ON_RETRACT                       0x00000100
#define D_EXO_API_CD_OPTION_INQUIRE_TOP_OF_FOR_ON_INIT                        0x00000200 
#define D_EXO_API_CD_OPTION_ENABLE_AUTOMATIC_RECOVERY_ON_INIT                 0x00000400 // Obsolete since Firmware R4-V2.00a 
#define D_EXO_API_CD_OPTION_INVALIDATE_PAPER_ON_INIT                          0x00000800
#define D_EXO_API_CD_OPTION_ADD_CUSTOMER_FLAG_TO_USB_VERSION                  0x00001000
#define D_EXO_API_CD_OPTION_PRINT_WHOLE_OBJECTS_ONLY                          0x00002000
#define D_EXO_API_CD_OPTION_ASSIGN_MODEL_SPECIFIC_USB_DEVICE_ID               0x00004000
#define D_EXO_API_CD_OPTION_DISABLE_RETRACT_AT_PRINT_START                    0x00008000
#define D_EXO_API_CD_OPTION_PRE_PULL_FORM_BEFORE_PRINT                        0x00010000
#define D_EXO_API_CD_OPTION_USE_IMPROVED_BURN_TIME_DETERMINATION              0x00020000 // AVAILABLE SINCE FIRMWARE R4-V2.00a
#define D_EXO_API_CD_OPTION_DISABLE_AUTO_PRESENTER_REJECT_FEATURE             0x00040000 // AVAILABLE SINCE FIRMWARE R4-V2.00b
#define D_EXO_API_CD_OPTION_DISABLE_PAPER_REJECT_ON_PAPER_LOAD_FAIL           0x00080000 // AVAILABLE SINCE FIRMWARE R4-V2.00c 
#define D_EXO_API_CD_OPTION_DISABLE_SCAN_OUTSIDE_MARK_AREA                    0x00100000 // AVAILABLE SINCE FIRMWARE R5-V2.01a 

//-- Configuration data - Options (Extendo-XF Firmware S684034)
#define D_EXF_API_CD_OPTION_CUTTER_INIT_ON_POWER_ON                           0x00000001
#define D_EXF_API_CD_OPTION_ENABLE_MARK_CONTROL                               0x00000002
#define D_EXF_API_CD_OPTION_PRINT_NO_INFO_ON_LOAD                             0x00000004
#define D_EXF_API_CD_OPTION_DISABLE_CUT_ON_LOAD                               0x00000008
#define D_EXF_API_CD_OPTION_INVALIDATE_UNUSED_PAPER                           0x00000010
#define D_EXF_API_CD_OPTION_NO_PAPER_FEED_AFTER_LOAD                          0x00000020
#define D_EXF_API_CD_OPTION_DISABLE_RETRACT_ON_PAPER_END                      0x00000040
#define D_EXF_API_CD_OPTION_CONTINUE_PRINT_ON_PAPER_END                       0x00000080
#define D_EXF_API_CD_OPTION_INQUIRE_TOP_OF_FOR_ON_INIT                        0x00000200 
#define D_EXF_API_CD_OPTION_ADD_CUSTOMER_FLAG_TO_USB_VERSION                  0x00001000
#define D_EXF_API_CD_OPTION_PRINT_WHOLE_OBJECTS_ONLY                          0x00002000
#define D_EXF_API_CD_OPTION_ASSIGN_MODEL_SPECIFIC_USB_DEVICE_ID               0x00004000
#define D_EXF_API_CD_OPTION_DISABLE_RETRACT_AT_PRINT_START                    0x00008000
#define D_EXF_API_CD_OPTION_PRE_PULL_FORM_BEFORE_PRINT                        0x00010000
#define D_EXF_API_CD_OPTION_DISABLE_AUTO_PRESENTER_REJECT_FEATURE             0x00040000 
#define D_EXF_API_CD_OPTION_DISABLE_PAPER_REJECT_ON_PAPER_LOAD_FAIL           0x00080000 
#define D_EXF_API_CD_OPTION_DISABLE_SCAN_OUTSIDE_MARK_AREA                    0x00100000 

//-- Sensor adjustment values (only available if sensor current adjustment feature is available)
#define D_EXO_API_CD_SCA_SETTING_DYNAMIC_DURING_INITIAL_PAPER_LOAD_ONLY       0          // Single dynamic adjustment (if possible)
#define D_EXO_API_CD_SCA_SETTING_FIXED_2BIT_10_0_mA                           1          // Sensor current fixed to 10,0 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_2BIT_15_0_mA                           2          // Sensor current fixed to 15,0 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_2BIT_20_0_mA                           3          // Sensor current fixed to 20,0 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_2BIT_24_0_mA                           4          // Sensor current fixed to 24,0 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_09_6_mA                           1          // Sensor current fixed to  9,6 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_10_8_mA                           2          // Sensor current fixed to 10,8 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_11_8_mA                           3          // Sensor current fixed to 11,8 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_13_1_mA                           4          // Sensor current fixed to 13,1 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_14_3_mA                           5          // Sensor current fixed to 14,3 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_15_6_mA                           6          // Sensor current fixed to 15,6 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_16_6_mA                           7          // Sensor current fixed to 16,6 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_17_8_mA                           8          // Sensor current fixed to 17,8 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_18_0_mA                           9          // Sensor current fixed to 18,0 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_19_3_mA                           10         // Sensor current fixed to 19,3 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_20_2_mA                           11         // Sensor current fixed to 20,2 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_21_5_mA                           12         // Sensor current fixed to 21,5 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_22_7_mA                           13         // Sensor current fixed to 22,7 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_23_9_mA                           14         // Sensor current fixed to 23,9 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_24_9_mA                           15         // Sensor current fixed to 24,9 mA
#define D_EXO_API_CD_SCA_SETTING_FIXED_4BIT_26_1_mA                           16         // Sensor current fixed to 26,1 mA
#define D_EXO_API_CD_SCA_SETTING_DYNAMIC_DURING_EVERY_PAPER_LOAD              255        // Default dynamic adjustment (if possible)

//-- Configuration Data => Font Sets (Extendo-XF Firmware S684034)
#define D_EXF_API_CD_FONT_SET_STANDARD                                        0
#define D_EXF_API_CD_FONT_SET_FACTORY                                         1
#define D_EXF_API_CD_FONT_SET_CUSTOMER                                        2

//-- Configuration Data => Character Sets (Extendo-XF Firmware S684034)
#define D_EXF_API_CD_CHAR_SET_0                                               0
#define D_EXF_API_CD_CHAR_SET_1                                               1
#define D_EXF_API_CD_CHAR_SET_2                                               2
#define D_EXF_API_CD_CHAR_SET_3                                               3
#define D_EXF_API_CD_CHAR_SET_4                                               4
#define D_EXF_API_CD_CHAR_SET_5                                               5
#define D_EXF_API_CD_CHAR_SET_6                                               6
#define D_EXF_API_CD_CHAR_SET_7                                               7
#define D_EXF_API_CD_CHAR_SET_8                                               8
#define D_EXF_API_CD_CHAR_SET_9                                               9
#define D_EXF_API_CD_CHAR_SET_10                                              10
#define D_EXF_API_CD_CHAR_SET_11                                              11
#define D_EXF_API_CD_CHAR_SET_12                                              12
#define D_EXF_API_CD_CHAR_SET_13                                              13
#define D_EXF_API_CD_CHAR_SET_14                                              14
#define D_EXF_API_CD_CHAR_SET_15                                              15

//-- Configuration Data => Character Dimensions (Extendo-XF Firmware S684034)
#define D_EXF_API_CD_CHAR_DIM_08X16                                           0      
#define D_EXF_API_CD_CHAR_DIM_12X24                                           1      
#define D_EXF_API_CD_CHAR_DIM_16X32                                           2
#define D_EXF_API_CD_CHAR_DIM_20X40                                           3

//-- Configuration Data => Character Encoding (Extendo-XF Firmware S684034)
#define D_EXF_API_CD_CHARACTER_ENCODING_CODEPAGE                              0
#define D_EXF_API_CD_CHARACTER_ENCODING_UTF8                                  1

//-- Configuration Data => Code Pages (Extendo-XF Firmware S684034)
#define D_EXF_API_CD_CP_CP00437_ENGLISH                                       0
#define D_EXF_API_CD_CP_CP00850_WESTERN_EUROPE                                2
#define D_EXF_API_CD_CP_CP00860_PORTUGUESE                                    3
#define D_EXF_API_CD_CP_CP00863_FRENCH_CANADA                                 4
#define D_EXF_API_CD_CP_CP00865_NORDIC                                        5
#define D_EXF_API_CD_CP_CP00852_CENTRAL_EUROPE                                6      
#define D_EXF_API_CD_CP_CP00866_RUSSIAN                                       7 
#define D_EXF_API_CD_CP_CP00857_TURKISH                                       8
#define D_EXF_API_CD_CP_CP01252_LATIN1                                        9
#define D_EXF_API_CD_CP_CP00858_WESTERN_EUROPE_EURO                           10
#define D_EXF_API_CD_CP_CP00862_HEBREW                                        11
#define D_EXF_API_CD_CP_CP01250_EASTERN_EUROPE                                12
#define D_EXF_API_CD_CP_CP01253_GREEK                                         13
#define D_EXF_API_CD_CP_CP01254_TURKISH                                       14
#define D_EXF_API_CD_CP_CP00861_ICELANDIC                                     15
#define D_EXF_API_CD_CP_CP01251_CYRILLIC                                      16
#define D_EXF_API_CD_CP_CP05000_MULTI_LANGUAGE                                17
#define D_EXF_API_CD_CP_ISO8859_1_LATIN_1_WESTERN_EUROPEAN                    30
#define D_EXF_API_CD_CP_ISO8859_2_LATIN_2_CENTRAL_EUROPEAN                    31
#define D_EXF_API_CD_CP_ISO8859_3_LATIN_3_SOUTH_EUROPEAN                      32
#define D_EXF_API_CD_CP_ISO8859_4_LATIN_4_NORTH_EUROPEAN                      33
#define D_EXF_API_CD_CP_ISO8859_5_LATIN_CYRILLIC                              34
#define D_EXF_API_CD_CP_ISO8859_6_LATIN_ARABIC                                35
#define D_EXF_API_CD_CP_ISO8859_7_LATIN_GREEK                                 36
#define D_EXF_API_CD_CP_ISO8859_8_LATIN_HEBREW                                37
#define D_EXF_API_CD_CP_ISO8859_9_LATIN_5_TURKISCH                            38
#define D_EXF_API_CD_CP_ISO8859_10_LATIN_6_NORDIC                             39
#define D_EXF_API_CD_CP_ISO8859_11_LATIN_THAI                                 40
#define D_EXF_API_CD_CP_ISO8859_13_LATIN_7_BALTIC_RIM                         41
#define D_EXF_API_CD_CP_ISO8859_14_LATIN_8_CELTIC                             42
#define D_EXF_API_CD_CP_ISO8859_15_LATIN_9_WESTERN_EUROPEAN                   43
#define D_EXF_API_CD_CP_ISO8859_16_LATIN_10_SOUTH_EASTERN_EUROPEAN            44
#define D_EXF_API_CD_CP_GB18030_CHINESE_NATIONAL_STANDARD                     60
#define D_EXF_API_CD_CP_GB2312_SIMPLIFIED_CHINESE                             61
#define D_EXF_API_CD_CP_KS5601_UNIFIED_HANGEUL                                62
#define D_EXF_API_CD_CP_BIG5_TRADITIONAL_CHINESE                              63

//-- Configuration Data => Text Attributes (Extendo-XF Firmware S684034)
#define D_EXF_API_CD_TEXT_ATTR_HOR_MAGNIFICATION                              0x0000000F      // Horizontal magnification mask
#define D_EXF_API_CD_TEXT_ATTR_HOR_MAGNIFICATION_1X                           0x00000000      // Horizontal magnification => OFF
#define D_EXF_API_CD_TEXT_ATTR_HOR_MAGNIFICATION_2X                           0x00000001      // Horizontal magnification => 2x 
#define D_EXF_API_CD_TEXT_ATTR_HOR_MAGNIFICATION_3X                           0x00000002      // Horizontal magnification => 3x 
#define D_EXF_API_CD_TEXT_ATTR_HOR_MAGNIFICATION_4X                           0x00000003      // Horizontal magnification => 4x 
#define D_EXF_API_CD_TEXT_ATTR_VER_MAGNIFICATION                              0x000000F0      // Vertical magnification mask
#define D_EXF_API_CD_TEXT_ATTR_VER_MAGNIFICATION_1X                           0x00000000      // Vertical magnification => OFF 
#define D_EXF_API_CD_TEXT_ATTR_VER_MAGNIFICATION_2X                           0x00000010      // Vertical magnification => 2x 
#define D_EXF_API_CD_TEXT_ATTR_VER_MAGNIFICATION_3X                           0x00000020      // Vertical magnification => 3x 
#define D_EXF_API_CD_TEXT_ATTR_VER_MAGNIFICATION_4X                           0x00000030      // Vertical magnification => 4x 
#define D_EXF_API_CD_TEXT_ATTR_UNDERLINE                                      0x00000F00      // Underline mask
#define D_EXF_API_CD_TEXT_ATTR_UNDERLINE_0                                    0x00000000      // Underline => 0 dot line
#define D_EXF_API_CD_TEXT_ATTR_UNDERLINE_1                                    0x00000100      // Underline => 1 dot line
#define D_EXF_API_CD_TEXT_ATTR_UNDERLINE_2                                    0x00000200      // Underline => 2 dot lines 
#define D_EXF_API_CD_TEXT_ATTR_UNDERLINE_3                                    0x00000300      // Underline => 3 dot lines 
#define D_EXF_API_CD_TEXT_ATTR_INVERSE                                        0x00001000      // Inverse
#define D_EXF_API_CD_TEXT_ATTR_BOLD                                           0x00002000      // Bold
#define D_EXF_API_CD_TEXT_ATTR_DOUBLE_WIDTH                                   0x00004000      // Double width
#define D_EXF_API_CD_TEXT_ATTR_DOUBLE_HEIGHT                                  0x00008000      // Double height
#define D_EXF_API_CD_TEXT_ATTR_ITALIC                                         0x00010000      // Italic

//-- Boolean macros
#define D_EXO_TRUE  1
#define D_EXO_FALSE 0

//-- Min max macros
#define M_EXO_MIN(x,y) (x>y ? y : x)
#define M_EXO_MAX(x,y) (x>y ? x : y)

//-- Buffer sizes
#define D_EXO_MAX_CONFIG_SIZE    255

//-- HEX conversion macros 
#define M_EXO_ISHEX(C)                                 ((((C) >= '0') && ((C) <= '9')) || ((toupper(C) >= 'A') && (toupper(C) <= 'F')))
#define M_EXO_HEX2VAL(C)                               ((((C) >= '0') && ((C) <= '9')) ? (C) - '0' : 10 + toupper(C) - 'A')
#define M_EXO_VAL2HEX(V)                               ((((V) >= 0) && ((V) <= 9)) ? '0' + (V) : 'A' + (V) - 10)

//-- Time macros
#define M_EXO_SLEEP_MS(X)                              usleep((unsigned long) (X)*1000)

//-- Synchronization
#define D_EXO_SYNCHRONIZATION_PACKET_OK                0x01010101                        

/*
** TYPES
*/

//-- Printer packet header (precedes printer status/info/etc. packets)
#pragma pack(1) 
typedef struct   
{ unsigned char                     ubSync1;                               // First synchronization byte (must be 0x1B)
   unsigned char                    ubSync2;                               // Second synchronization byte (must be 0xFF)
   unsigned char                    ubId;                                  // Identification of packet 
   unsigned char                    ubLength;                              // Length of header and data part
} TS_EXO_API_PCKT_HEADER;                                                  // Packet header
#pragma pack() 

//-- Printer status packet (identifier in preceding header is 2)
#pragma pack(1) 
typedef struct   
{ uint32_t                          dwStatus;                              // Printer status dword as in D_EXO_API_PRINTER_STATUS_PSDW_...
   unsigned char                    ubPrintJobId;                          // Id of print job
   short                            siPrintHeadTempC;                      // Printhead temperature in degrees celcius                          (ADC#02)
   unsigned short                   wBoardVoltageTv;                       // Board voltage in 1/10 volt
   unsigned char                    ubState;                               // Status of paper control
   unsigned short                   wError;                                // Printer error word as in D_EXO_API_PRINTER_STATUS_PEW_...
   unsigned char                    ubReserved[2];                         // Reserved bytes
} TS_EXO_API_PCKT_PRINTER_STATUS;                                          // Printer status packet 
#pragma pack() 

//-- Printer info packet (identifier in preceding header is 3)
#pragma pack(1) 
typedef struct 
{ unsigned char                     ubProtocolRevision;                    // Revision of communication protocol (see D_EXO_API_PCKT_REVISION)
   char                             szPartNr[16+1];                        // Part number, inluding 0 terminator
   char                             szSerialNr[16+1];                      // Serial number, inluding 0 terminator
   char                             szDateOfManuf[16+1];                   // Date of manufacturing, inluding 0 terminator
   uint32_t                         dwSysConfig;                           // Configuration code
   uint32_t                         dwSysSensor;                           // Sensor code
   uint32_t                         dwSysFeature;                          // Feature code
   char                             szFwPartNr[16+1];                      // Part number, inluding 0 terminator 
   char                             szFwVersion[10+1];                     // Firmware version  Rx-Vy.zzb
   char                             szFwDate[12+1];                        // Firmware date     Mmm dd yyyy (ANSI standard).
   uint32_t                         dwFwPatch;                             // Firmware patch code
} TS_EXO_API_PCKT_PRINTER_INFO;                                            // Printer info packet
#pragma pack() 

//-- Printer sensor packet (identifier in preceding header is 4)
#pragma pack(1) 
typedef struct
{ unsigned char                     ubProtocolRevision;                    // Revision of communication protocol
   unsigned char                    ubAdcMeasurement[16];                  // ADC channel #00..#15 measurements in domain 0=0V..255=5V                (ADC#XX)
   unsigned short                   wDigitalInput;                         // DIN digital inputs bits #00..#15                                        (DIN#XX)
   short                            siPrintHeadTempC;                      // Printhead temperature in degrees celcius                                (ADC#02)
   unsigned short                   wBoardVoltageTv;                       // Board voltage in tenths of volt                                         (ADC#01)
   unsigned char                    ubPaperPreEndSensorState;              // State byte for paper pre end sensor   (D_EXO_API_PRINTER_SENSOR_PSS_..) (ADC#03)
   unsigned char                    ubPaperEntrySensor1State;              // State byte for paper entry sensor #1  (D_EXO_API_PRINTER_SENSOR_PSS_..) (ADC#04)
   unsigned char                    ubPaperEntrySensor2State;              // State byte for paper entry sensor #2  (D_EXO_API_PRINTER_SENSOR_PSS_..) (ADC#05)
   unsigned char                    ubHeadUpSensorState;                   // State byte for head up sensor         (D_EXO_API_PRINTER_SENSOR_PSS_..) (DIN#00)
   unsigned char                    ubCutterSensor1State;                  // State byte for cutter sensor #1       (D_EXO_API_PRINTER_SENSOR_PSS_..) (DIN#01)
   unsigned char                    ubCutterSensor2State;                  // State byte for cutter sensor #2       (D_EXO_API_PRINTER_SENSOR_PSS_..) (DIN#02)
   unsigned char                    ubPaperExitSensorState;                // State byte for paper exit sensor      (D_EXO_API_PRINTER_SENSOR_PSS_..) (ADC#00)
   unsigned char                    ubAuxiliarySensorState;                // State byte for auxiliary sensor       (D_EXO_API_PRINTER_SENSOR_PSS_..) (ADC#06)
   unsigned char                    ubPresenterSensor1State;               // State byte for presenter sensor #1    (D_EXO_API_PRINTER_SENSOR_PSS_..) (ADC#00)
   unsigned char                    ubPresenterSensor2State;               // State byte for presenter sensor #2    (D_EXO_API_PRINTER_SENSOR_PSS_..) (ADC#06)
   unsigned char                    ubReserved[2];                         // Reserved bytes, must be 0
} TS_EXO_API_PCKT_PRINTER_SENSOR;                                          // Printer sensor packet
#pragma pack() 

//-- Printer flash packet (identifier in preceding header is 5)
#pragma pack(1) 
typedef struct 
{ unsigned char                     ubProtocolRevision;                    // Revision of communication protocol
   char                             szInfo[6][20+1];                       // Flash area info, indexed by D_EXO_API_PRINTER_FLASH_FA_INDEX_..., inluding 0 terminator
   unsigned short                   wCrc16[6][2];                          // Flash area 16 bit checksum, indexed by D_EXO_API_PRINTER_FLASH_FA_INDEX_..., 0=stored and 1=calculated
   unsigned char                    ubState[6];                            // Flash area state, indexed by D_EXO_API_PRINTER_FLASH_FA_INDEX_..., 
   unsigned char                    ubReserved[6];                         // Reserved bytes, must be 0
} TS_EXO_API_PCKT_PRINTER_FLASH;                                           // Printer flash packet
#pragma pack() 

//-- Presenter status packet (identifier in preceding header is 6)
#pragma pack(1) 
typedef struct  
{ unsigned char                     ubProtocolRevision;                    // Revision of communication protocol
   unsigned char                    ubType;                                // Type of presenter (0 is NONE)
   unsigned short                   wStatus;                               // Presenter status word bits as in D_EXO_API_PRESENTER_STATUS_PSW_...
   unsigned char                    ubState;                               // State of presenter control (idle/loading/presenting/rejecting, etc.)
   unsigned char                    ubNrOfLoadedPages;                     // Total number of currently loaded  (bundled) printouts
   uint32_t                         dwNrOfProcessedPages;                  // Total number of processed pages since start / power on
   uint32_t                         dwNrOfEjectedPages;                    // Total number of ejected   pages since start / power on
   uint32_t                         dwNrOfRejectedPages;                   // Total number of rejected  pages since start / power on
   unsigned char                    ubReserved[8];                         // Reserved bytes, must be 0
} TS_EXO_API_PCKT_PRESENTER_STATUS;                                        // Presenter status packet
#pragma pack() 

//-- Life cycle data packet
#pragma pack(1) 
typedef struct 
{ unsigned char                     ubProtocolRevision;                    // Revision of communication protocol
   uint32_t                         dwTimePoweredS;                        // Total time that the printer is powered in sec
   uint32_t                         dwDistanceSteppedMm;                   // Total distance stepped in mm
   uint32_t                         dwDistanceBurnedMm;                    // Total distance burned in mm
   uint32_t                         dwPaperTransportedMm;                  // Total length of transported paper in mm
   uint32_t                         dwNrOfSystemStarts;                    // Total number of system starts
   uint32_t                         dwNrOfHardResets;                      // Total number of hardware resets 
   uint32_t                         dwNrOfPaperCuts;                       // Total number of paper cuts 
   unsigned short                   wNrOfPaperLoads;                       // Total number of paper loads 
   unsigned char                    ubReserved[8];                         // Reserved bytes (2 Dunsigned shorts, zero initialized)
}  TS_EXO_API_PCKT_PRINTER_LIFE_CYCLE;                                     // Life cycle packet
#pragma pack() 

//-- Configuration data structure for normal eXtendo firmware with P/N S684014 
#pragma pack(1) // Byte aligned
typedef struct 
{ unsigned short                    wRevisionNumber;                       // Structure revision number 
   unsigned short                   wSizeOfBlock;                          // Size of this structure
   char                             szCustSerialNr[16+1];                  // Customer-serial number, zero is unused
   char                             szCustPartNr[16+1];                    // Customer-part number, zero is unused
   unsigned char                    ubCustFlags;                           // Customer specific flags
   unsigned char                    ubBaudRate;                            // Baudrate matching one of D_EXO_API_CD_RS232_BAUDRATE_...
   unsigned char                    ubParityOnOff;                         // Parity matching one of D_EXO_API_CD_RS232_PARITY_... (on/off)
   unsigned char                    ubParity;                              // Parity matching one of D_EXO_API_CD_RS232_PARITY_... (even/odd, if on)
   unsigned char                    ubDataBits;                            // Data bits matching one of D_EXO_API_CD_RS232_DATABIT_... (7/ 8)
   unsigned char                    ubStopBits;                            // Stop bits matching one of D_EXO_API_CD_RS232_STOPBIT_... (1/2)
   unsigned char                    ubFlowControl;                         // Flow control matching one of D_EXO_API_CD_RS232_FLOWCNTL_... (SW/HW/NO)
   unsigned char                    ubBlockHostTxOnPaperOut;               // Flag indicating disable host TX on paper out (RS232-HW/SW flow control only)
   unsigned char                    ubWideFont;                            // Enumeration indicating wide font set enabled if non zero
   unsigned short                   wPrintSpeedTmms;                       // Print speed (default) in 1/10 mm/sec
   char                             cPrintDensity;                         // Print density (default) in percent
   unsigned char                    ubDhisBurnFac;                         // Dot history burn factor in percent of total burn time (0 = dot history disabled)
   unsigned char                    ubMultiStrobeFac;                      // Multi strobe counter, zero is undefined
   char                             asbBurnTimeCorr[11];                   // Burn time corrections in us for -30°..+70°C with 10°C interval
   uint32_t                         dwOptions;                             // Options bitmap as defined in D_EXO_API_CD_OPTION_...
   uint32_t                         dwFormLengthYm;                        // Maximum form length in micrometer
   uint32_t                         dwMarkCutDisYm;                        // Distance from end of black/hole mark to cut position in micrometer
   unsigned short                   wMarkSizeYm;                           // Size of a black/hole mark in micrometer
   unsigned char                    ubPaperParkPos;                        // Paper park position 0 or 1
   uint32_t                         dwPaperLoadFeedYm;                     // Feed length after paper load in micrometer
   unsigned short                   wCustSensor;                           // Customer sensors as in D_EXO_API_CD_CS_.... 
   unsigned short                   wPaperRemovalPeriodSec;                // Duraation of the paper removal period after partial cut in seconds
   unsigned short                   wPaperEndDetectDisYm;                  // Minimum distance required before detecting paper end in micrometer
   uint32_t                         dwPaperPresentationOffsetYm;           // Distance in which paper if forwarded after cut in micrometer (from firwmare 1.02)
   short                            iMarkCutCorrectionYm;                  // Correction of mark cut position in micrometer (from firwmare 1.02)
   short                            iPrintCorrectionYm;                    // Correction of print start position in micrometer (from firwmare 1.02)
   unsigned char                    ubMarkThresholdCorrection;             // Mark threshold correction in 10% (1/10 volt) resolution (from firmware 1.03j)
   unsigned short                   wPaperLoadSpeedTmms;                   // Paper load    speed in 1/10 mm/sec (from firmware 2.00a)
   unsigned short                   wPaperReverseSpeedTmms;                // Paper reverse speed in 1/10 mm/sec (from firmware 2.00a)
   unsigned short                   wPaperAdvanceSpeedTmms;                // Paper advance speed in 1/10 mm/sec (from firmware 2.00a)
   unsigned char                    ubScaPaperEntrySensor;                 // Paper entry sensor current adjustment (SCA) as in D_EXO_API_CD_SCA_... (from firmware 2.01g)
   unsigned char                    ubScaPaperExitSensor;                  // Printer exit sensor current adjustment (SCA) as in D_EXO_API_CD_SCA_... (from firmware 2.01g)
   unsigned char                    ubScaPresenterExitSensor;              // Presenter exit sensor current adjustment (SCA) as in D_EXO_API_CD_SCA_... (from firmware 2.01g)
   unsigned char                    ubPreMarkValidationDistanceTmm;        // Pre-mark validation distance in 1/10 mm (from firmware 2.01i)
   unsigned char                    ubPostMarkValidationDistanceTmm;       // Post-mark validation distance in 1/10 mm (from firmware 2.01i  )
   unsigned char                    ubScaPaperEntrySensorDynamicMin;       // Minimum dynamic paper entry sensor current adjustment (SCA) as in D_EXO_SM_SCA_... (from firmware 2.01i)
   unsigned char                    ubScaPaperEntrySensorDynamicMax;       // Maximum dynamic paper entry sensor current adjustment (SCA) as in D_EXO_SM_SCA_... (from firmware 2.01i)
   unsigned char                    ubReserved[8];                         // Reserved byte 0-8 must be zero
   unsigned short                   wCrc16;                                // 16 Bit checksum of structure
} TS_EXO_API_CONFIG_DATA;                                                  // Extendo configuration data structure as uploaded to and stored in flash
#define D_EXO_API_CONFIG_DATA_REV 1                                        // Revision of configuration data
#pragma pack() 

//-- Configuration data structure for eXtendo-XF firmware with P/N S684034 
#pragma pack(1) // Byte aligned
typedef struct
{ unsigned short                    wRevisionNumber;                       // Structure revision number 
   unsigned short                   wSizeOfBlock;                          // Size of this structure
   char                             szCustDescription[32+1];               // Description, zero is unused
   unsigned char                    ubCustFlags;                           // Customer specific flags
   unsigned char                    ubBaudRate;                            // Baudrate matching one of D_EXO_SM_RS232_BAUDRATE_...
   unsigned char                    ubParityOnOff;                         // Parity matching one of D_EXO_SM_RS232_PARITY_... (on/off)
   unsigned char                    ubParity;                              // Parity matching one of D_EXO_SM_RS232_PARITY_... (even/odd, if on)
   unsigned char                    ubDataBits;                            // Data bits matching one of D_EXO_SM_RS232_DATABIT_... (7/8)
   unsigned char                    ubStopBits;                            // Stop bits matching one of D_EXO_SM_RS232_STOPBIT_... (1/2)
   unsigned char                    ubFlowControl;                         // Flow control matching one of D_EXO_SM_RS232_FLOWCNTL_... (SW/HW/NO)
   unsigned char                    ubBlockHostTxOnPaperOut;               // Flag indicating disable host TX on paper out (RS232-HW/SW flow control only)
   unsigned short                   wPrintSpeedTmms;                       // Default print speed in 1/10 mm/s
   char                             cPrintDensityPrcnt;                    // Default print density in domain [-100%..+100%]
   unsigned char                    ubDhisLevel;                           // Dot history level (0 = dot history disabled)
   unsigned char                    ubDhisBurnFac;                         // Dot history burn factor in percent of total burn time (0 = dot history disabled)
   unsigned char                    ubMultiStrobeFac;                      // Multi strobe counter, zero is single, one is double
   uint32_t                         dwOptions;                             // Options bitmap as defined in D_EXF_API_CD_OPTION_...
   unsigned short                   wFormLengthTmm;                        // Maximum form length in 1/10 mm
   unsigned short                   wMarkCutDisTmm;                        // Distance from end of black/hole mark to cut position in 1/10 mm
   unsigned short                   wMarkLengthTmm;                        // Length of a black/hole mark in 1/10 mm
   unsigned char                    ubPaperParkPos;                        // Paper park position 0 or 1
   unsigned short                   wPaperLoadFeedTmm;                     // Paper feed on paper load in 1/10 mm
   unsigned short                   wCustSensor;                           // Customer sensors as in D_EXO_CS_.... 
   unsigned short                   wPaperRemovalPeriodSec;                // Duration of the paper removal period after partial cut in seconds
   unsigned short                   wPaperEndDetectDisTmm;                 // Minimum distance required before detecting paper end in 1/10 mm
   unsigned short                   wPaperPresentationOffsetTmm;           // Distance to go forward/backward after cut in 1/10 mm
   short                            iMarkCutCorrectionTmm;                 // Mark cut correction in 1/10 mm
   short                            iPrintCorrectionTmm;                   // Print start correction in 1/10 mm
   unsigned char                    ubMarkThresholdCorrection;             // Mark threshold correction in 10% (1/10 volt) resolution
   unsigned short                   wPaperLoadSpeedTmms;                   // Paper load    speed in 1/10 mm/sec
   unsigned short                   wPaperReverseSpeedTmms;                // Paper reverse speed in 1/10 mm/sec
   unsigned short                   wPaperAdvanceSpeedTmms;                // Paper advance speed in 1/10 mm/sec
   unsigned char                    ubScaPaperEntrySensor;                 // Paper entry sensor current adjustment (SCA) as in D_EXO_SM_SCA_...
   unsigned char                    ubScaPaperExitSensor;                  // Printer exit sensor current adjustment (SCA) as in D_EXO_SM_SCA_... 
   unsigned char                    ubScaPresenterExitSensor;              // Presenter exit sensor current adjustment (SCA) as in D_EXO_SM_SCA_... 
   unsigned char                    ubPreMarkValidationDistanceTmm;        // Pre-mark validation distance in 1/10 mm 
   unsigned char                    ubPostMarkValidationDistanceTmm;       // Post-mark validation distance in 1/10 mm 
   unsigned char                    ubScaPaperEntrySensorDynamicMin;       // Minimum dynamic paper entry sensor current adjustment as in D_EXO_SM_SCA_... 
   unsigned char                    ubScaPaperEntrySensorDynamicMax;       // Maximum dynamic paper entry sensor current adjustment as in D_EXO_SM_SCA_... 
   unsigned char                    ubFontSet;                             // Default font set matching one or more of D_EXF_API_CD_FONT_SET_...
   unsigned char                    ubCharSet;                             // Default character set matching one or more of D_EXF_API_CD_char_SET_...
   unsigned char                    ubCharDim;                             // Default character dimension matching one or more of D_EXF_API_CD_char_DIM_...
   unsigned char                    ubCharacterEncodingMode;               // Default character encoding mode, matching one of D_EXF_API_CD_charACTER_ENCODING_...
   unsigned char                    ubCodePage;                            // Default code page matching one or more of D_EXF_API_CD_CODE_PAGE_...
   uint32_t                         dwTextAttributes;                      // Default text attributes matching one or more of D_EXF_API_CD_TEXT_ATTR_...
   unsigned char                    ubCharSpaceDots;                       // Default character space in dots
   unsigned char                    ubLineSpaceDots;                       // Default line space in dots 
   unsigned short                   wLeftMarginTmm;                        // Default left margin in 1/10 mm
   unsigned short                   wPaperOffsetTmm;                       // Paper offset in 1/10 mm 
   unsigned short                   wPaperWidthTmm;                        // Paper width in 1/10 mm  (0 is default)
   unsigned char                    aubReserved[64];                       // Reserved bytes (must be zero)
   unsigned short                   wCrc16;                                // 16 Bit checksum of structure
} TS_EXF_API_CONFIG_DATA;                                                  // Configuration data structure as uploaded to and stored in flash
#define D_EXF_API_CONFIG_DATA_REV 1                                        // Revision of configuration data
#pragma pack() 



/*
** VARIABLES
*/

/*
** FUNCTIONS
*/

//-- Inquiry function 
char *         exo_api_get_library_info      (void);

//-- Printer communication interface
intptr_t       exo_api_printer_open          (char * port);       // e.g. port = "/dev/usb/lp0" or port = "/dev/bus/usb/003/002"
int            exo_api_printer_set_serial    (intptr_t printer, unsigned long baud_rate, unsigned char data_bits, unsigned char parity, unsigned char stop_bits);
void           exo_api_printer_close         (intptr_t printer);
int            exo_api_printer_write         (intptr_t printer, unsigned char * data, int size, unsigned long timeout_ms);
int            exo_api_printer_read          (intptr_t printer, unsigned char * data, int size, unsigned long timeout_ms);
int            exo_api_printer_clear         (intptr_t printer);

#ifdef __cplusplus
}
#endif

#endif // __EXO_API_H







